from typing import Any

__all__ = ["Tp"]


def Tp(mixin_specs: list[type | tuple], prototype: Any) -> type:
    base_cls = type(prototype)

    mixin_bases = [spec[0] if isinstance(spec, tuple) else spec for spec in mixin_specs]
    name = base_cls.__name__ + "".join(c.__name__ for c in mixin_bases)
    bases = (base_cls,) + tuple(mixin_bases)

    def __new__(cls, *args, **kwargs):
        return base_cls.__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        base_cls.__init__(self, *args, **kwargs)
        for spec in mixin_specs:
            if isinstance(spec, tuple):
                mixin_cls, *mixin_args = spec
                mixin_cls.__init__(self, *mixin_args)
            else:
                spec.__init__(self)

    NewType = type(name, bases, {"__new__": __new__, "__init__": __init__})

    for spec in mixin_specs:
        mixin_cls = spec[0] if isinstance(spec, tuple) else spec
        mixin_args = spec[1:] if isinstance(spec, tuple) else []

        if hasattr(mixin_cls, "validate_type"):
            try:
                mixin_cls.validate_type(prototype, *mixin_args)
            except Exception as e:
                raise TypeError(
                    f"Validation failed for '{mixin_cls.__name__}' on base class '{base_cls.__name__}'. "
                    f"The base type does not satisfy the required laws."
                ) from e

    return NewType



from functools import reduce, partial
from typing import TypeVar
from inspect import signature, _empty

__all__ = ["identity", "assert_and", "expr", "compose", "curry", "Filter", "State"]


def identity(x):
    return x


T = TypeVar("T")


def assert_and(cond: bool, expr: T, msg: str | None = "Assertion failed") -> T:
    """Works like assert but returns expr when cond is True"""
    assert cond, msg
    return expr


def expr(*args):
    """Evaluates and returns the last argument"""
    return args[-1]


def compose(*funcs):
    """Composes multiple functions into a single function"""
    return reduce(lambda f, g: lambda x: f(g(x)), funcs, identity)


# from https://stackoverflow.com/a/78149460
# TODO: since it uses inspect, it doesn't work with built-in functions
def curry(f):
    def inner(*args, **kwds):
        new_f = partial(f, *args, **kwds)
        params = signature(new_f, follow_wrapped=True).parameters
        if all(params[p].default != _empty for p in params):
            return new_f()
        else:
            return curry(new_f)

    return inner


def cache(func):
    _cache = {}

    def wrapper(*args):
        if args not in _cache:
            result = func(*args)
            _cache[args] = result
            return result
        else:
            return _cache[args]

    return wrapper


class Filter:
    def __init__(self, func):
        self.func = func

    def __rmatmul__(self, iterable):
        return type(iterable)(filter(self.func, iterable))


class State:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def update(self, **kwargs) -> "State":
        new_state = self.__dict__.copy()
        return State(**{**new_state, **kwargs})

    def __repr__(self):
        return (
            "Ctx {\n"
            + "\n".join(f"  {k}={v}" for k, v in self.__dict__.items())
            + "\n}"
        )



from .type_class import Tp
from .type_classes import Semigroup, Functor
from .utils import identity, assert_and, expr, compose, curry, Filter, State

lst = Tp([Functor, Semigroup], [1, 2, 3])

# fmt: off
__all__ = [
    "Semigroup", "Functor",
    "identity", "assert_and", "expr", "compose", "curry",
    "Filter", "State",
    "lst"
]
# fmt: on



from typing import Any
from collections.abc import Callable

__all__ = ["Semigroup", "Functor"]


def _call_op(a: Any, b: Any, op: str) -> Any:
    return getattr(a, f"__{op}__")(b)


class Semigroup:
    """
    A semigroup should satify:
      * Associativity
      * Unit
      * Multiplication

    Default operation is '+' so that it works with numbers, strings, lists, tuples, etc.
    """

    @staticmethod
    def validate_type(prototype: Any, op: str = "add") -> None:
        Type = type(prototype)
        ident = Type()

        assert _call_op(_call_op(ident, prototype, op), ident, op) == _call_op(
            ident, _call_op(prototype, ident, op), op
        ), f"Type does not satisfy associativity for op ${op}"
        assert (
            _call_op(ident, prototype, op) == prototype
            and _call_op(prototype, ident, op) == prototype
        ), "Type does not satisfy unit"
        assert _call_op(prototype, _call_op(prototype, prototype, op), op) == _call_op(
            _call_op(prototype, prototype, op), prototype, op
        ), f"Type does not satisfy multiplication for op ${op}"

    def __init__(self, op: str = "add") -> None:
        self.op = op

    def __add__(self, other: "Semigroup") -> "Semigroup":
        assert type(self) is type(other) and other.op == self.op
        return type(self)(_call_op(self, other, self.op))


class Functor:
    """
    A functor should satisfy:
      * Identity
      * Composition

    Works with lists, tuples, sets, etc.
    """

    @staticmethod
    def validate_type(prototype: Any) -> None:
        base_cls = type(prototype)
        assert (
            base_cls(map(lambda x: x, prototype)) == prototype
        ), "Type does not satisfy identity"

    def fmap(self, func: Callable) -> "Functor":
        return type(self)(map(func, self))  # pyright: ignore

    def __or__(self, func: Callable) -> "Functor":
        return self.fmap(func)
